//=============================================================================
// FireRingDemo.cpp by Frank Luna (C) 2005 All Rights Reserved.
//
// Demonstrates a ring of fire particle system.
//
// Controls: Use mouse to look and 'W', 'S', 'A', and 'D' keys to move.
//=============================================================================

#include "d3dApp.h"
#include "DirectInput.h"
#include <crtdbg.h>
#include "GfxStats.h"
#include <list>
#include <ctime>
#include "Terrain.h"
#include "Camera.h"
//#include "PSystem.h"



class FireRingDemo : public D3DApp
{
public:
	FireRingDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP);
	~FireRingDemo();

	bool checkDeviceCaps();
	void onLostDevice();
	void onResetDevice();
	void updateScene(float dt);
	void drawScene();
	void initScene();
	bool soft;
private:
	GfxStats* mGfxStats;
	Terrain*  mTerrain;


	D3DXMATRIX					matTranslation;
	D3DXMATRIX					matScale;
	D3DXMATRIX					matMultiply;

	D3DXMATRIX					psysWorld;
	ID3DXEffect					*skyFx;
	ID3DXEffect					*softParticleFx;
	//Deph map stuff:
	IDirect3DTexture9			*pDepthMap;
	IDirect3DSurface9			*pDepthMapSurf;
	IDirect3DSurface9			*pDepthMapZ;

	IDirect3DSurface9			*pOrigRT;
	IDirect3DSurface9			*pOrigZ;


	IDirect3DTexture9			*particleMap;
	IDirect3DTexture9			*skyMap;

	IDirect3DVertexBuffer9		*pVBQuad;
	IDirect3DVertexDeclaration9	*pDeclPos;

	std::vector<Mtrl> mSkyMtrls;
	std::vector<IDirect3DTexture9*> mSkyTextures;
	ID3DXMesh* mSkyMesh;


};


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance,
				   PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	srand(time(0));

	// Construct camera before application, since the application uses the camera.
	Camera camera;
	gCamera = &camera;

	FireRingDemo app(hInstance, "Fire Ring Demo", D3DDEVTYPE_HAL, D3DCREATE_HARDWARE_VERTEXPROCESSING);
	gd3dApp = &app;

	DirectInput di(DISCL_NONEXCLUSIVE|DISCL_FOREGROUND, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	gDInput = &di;

    return gd3dApp->run();
}

FireRingDemo::FireRingDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP)
: D3DApp(hInstance, winCaption, devType, requestedVP)
{
	if(!checkDeviceCaps())
	{
		MessageBox(0, "checkDeviceCaps() Failed", 0, 0);
		PostQuitMessage(0);
	}

	InitAllVertexDeclarations();
	initScene();
	mGfxStats = new GfxStats();
	
	mTerrain = new Terrain(257, 257, 2.0f, 2.0f, 
		"heightmap1_257.raw",  
		"mud.dds",
		"stone.dds",
		"snow.dds",
		"blend_hm1.dds",
		0.4f, 0.0f);

	D3DXVECTOR3 toSun(1.0f, 1.0f, 1.0f);
	D3DXVec3Normalize(&toSun, &toSun);
	mTerrain->setDirToSunW(toSun);

	// Initialize camera.
	gCamera->pos() = D3DXVECTOR3(55.0f, 50.0f, 25.0f);
	gCamera->setSpeed(40.0f);

	// Initialize the particle system.

	D3DXMatrixTranslation(&psysWorld, 55.0f, 55.0f, 55.0f);
	AABB psysBox;
	psysBox.minPt = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	psysBox.maxPt = D3DXVECTOR3( 0.0f, 0.0f,  0.0f);

	
	mGfxStats->addVertices(mTerrain->getNumVertices());
	mGfxStats->addTriangles(mTerrain->getNumTriangles());

	onResetDevice();
}

FireRingDemo::~FireRingDemo()
{
	delete mGfxStats;
	delete mTerrain;
	SAFE_RELEASE(skyMap);
	SAFE_RELEASE(particleMap);
	SAFE_RELEASE(pDepthMap);
	SAFE_RELEASE(pDepthMapSurf);
	SAFE_RELEASE(pDepthMapZ);
	SAFE_RELEASE(pDeclPos);
	SAFE_RELEASE(pVBQuad);
	SAFE_RELEASE(pOrigRT);
	SAFE_RELEASE(pOrigZ);
	SAFE_RELEASE(softParticleFx);
	SAFE_RELEASE(skyFx);
	SAFE_RELEASE(mSkyMesh);
	mSkyMtrls.clear();
	mSkyTextures.clear();

	DestroyAllVertexDeclarations();
}

bool FireRingDemo::checkDeviceCaps()
{
	D3DCAPS9 caps;
	HR(gd3dDevice->GetDeviceCaps(&caps));

	// Check for vertex shader version 2.0 support.
	if( caps.VertexShaderVersion < D3DVS_VERSION(2, 0) )
		return false;

	// Check for pixel shader version 2.0 support.
	if( caps.PixelShaderVersion < D3DPS_VERSION(2, 0) )
		return false;

	return true;
}

void FireRingDemo::onLostDevice()
{
	mGfxStats->onLostDevice();
	mTerrain->onLostDevice();

}

void FireRingDemo::onResetDevice()
{
	mGfxStats->onResetDevice();
	mTerrain->onResetDevice();

	// The aspect ratio depends on the backbuffer dimensions, which can 
	// possibly change after a reset.  So rebuild the projection matrix.
	float w = (float)md3dPP.BackBufferWidth;
	float h = (float)md3dPP.BackBufferHeight;
	gCamera->setLens(D3DX_PI * 0.25f, w/h, 0.01f, 50000.0f);
}

void FireRingDemo::updateScene(float dt)
{
	mGfxStats->update(dt);
	if(gDInput->keyDown(DIK_P))
	{	
		soft=!soft;
	}

	gDInput->poll();

	gCamera->update(dt, 0, 0);

			
	softParticleFx->SetBool( "soft", soft);		
}

void FireRingDemo::drawScene()
{

	// Clear the backbuffer and depth buffer.
	HR(gd3dDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xFF0FFFFF, 1.0f, 0));

	HR(gd3dDevice->BeginScene());

		UINT p;

		gd3dDevice->GetRenderTarget(0, &pOrigRT);
		gd3dDevice->GetDepthStencilSurface(&pOrigZ);

		gd3dDevice->SetRenderTarget(0, pDepthMapSurf);
		gd3dDevice->SetDepthStencilSurface(pDepthMapZ);

		gd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xF0FFFFFF, 1.0f, 0 );

			D3DXMatrixScaling( &matScale,1.1,1.1,1.1);	
			D3DXMatrixTranslation( &matTranslation,gCamera->pos().x, gCamera->pos().y-300.0, gCamera->pos().z);
			matMultiply = (matScale * matTranslation);
			gd3dDevice->SetTransform( D3DTS_WORLD, &matMultiply);

			skyFx->SetMatrix("matWVP", &(matMultiply*gCamera->viewProj()));
			skyFx->SetTechnique("drawDepth");
			skyFx->CommitChanges();
			skyFx->Begin(&p,0);
				skyFx->BeginPass(0);
					HR(mSkyMesh->DrawSubset(0));
				skyFx->EndPass();
			skyFx->End();

			mTerrain->drawDepth();

		gd3dDevice->SetRenderTarget(0, pOrigRT);
		gd3dDevice->SetDepthStencilSurface(pOrigZ);


		// Build light projective texture matrix.
		D3DXMATRIX lightView;
		D3DXVECTOR3 lightPosW(gCamera->pos().x, -800.0f, gCamera->pos().z+0.01f);
		D3DXVECTOR3 lightTargetW(gCamera->pos().x, 0.0, gCamera->pos().z);
		D3DXVECTOR3 lightUpW(0.0f, 1.0f, 0.0f);
		D3DXMatrixLookAtLH(&lightView, &lightPosW, &lightTargetW, &lightUpW);
		
		D3DXMATRIX lightLens;
		float lightFOV = D3DX_PI*0.30f;
		D3DXMatrixPerspectiveFovLH(&lightLens, lightFOV, 1.0f, 1.0f, 5000.0f);

		skyFx->SetMatrix("gLightWVP", &(matMultiply*lightView*lightLens));
		skyFx->SetMatrix("matWVP", &(matMultiply*gCamera->viewProj()));

		skyFx->SetTexture("gTex0", skyMap);
		skyFx->SetTechnique("DrawSky");
		skyFx->CommitChanges();
		skyFx->Begin(&p,0);
			skyFx->BeginPass(0);
				HR(mSkyMesh->DrawSubset(0));
			skyFx->EndPass();
		skyFx->End();


		mTerrain->draw();


		softParticleFx->SetValue("pozicija", D3DXVECTOR3(55.0f, 50.0f, 25.0f), sizeof(D3DXVECTOR3));
		softParticleFx->SetValue("gEyePosW", &gCamera->look(), sizeof(D3DXVECTOR3));

		softParticleFx->SetValue("luk",  &gCamera->look(), sizeof(D3DXVECTOR3));
		softParticleFx->SetValue("rajt", &gCamera->right(), sizeof(D3DXVECTOR3));
		softParticleFx->SetValue("ap",   &gCamera->up(), sizeof(D3DXVECTOR3));

		softParticleFx->SetMatrix("gViewProj", &gCamera->viewProj());
		softParticleFx->SetMatrix("matView", &gCamera->view());	
		softParticleFx->SetFloat( "sc", 12.5f);
		HR(softParticleFx->SetTexture("colorTex2", particleMap));
		HR(softParticleFx->SetTexture("colorTex1", pDepthMap));

		softParticleFx->SetTechnique("Billboard");
		softParticleFx->Begin(&p,0);
			softParticleFx->BeginPass(0);
				gd3dDevice->SetVertexDeclaration(pDeclPos);
				gd3dDevice->SetStreamSource(0, pVBQuad,  0,sizeof(D3DXVECTOR3));
				gd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP,0, 2);
			softParticleFx->EndPass();
		softParticleFx->End();


	char buffer[256];

	sprintf(buffer, "camx %1f, camy %1f, camz %1f", gCamera->pos().x, gCamera->pos().y,gCamera->pos().z);
	if(soft){strcat(buffer, "\nSoft Particles");}else{strcat(buffer, "\nHard Particles");}


		mGfxStats->display(buffer);

	HR(gd3dDevice->EndScene());

	// Present the backbuffer.
	HR(gd3dDevice->Present(0, 0, 0, 0));
}

void FireRingDemo::initScene(void)
{
	LPD3DXBUFFER pBufferErrors = NULL;
	
	if (FAILED(D3DXCreateEffectFromFile(gd3dDevice, TEXT("particles.fx"), 0, 0, 0, 0, &softParticleFx, &pBufferErrors)))
	{
		if (pBufferErrors)
		{
			LPVOID pCompilErrors = pBufferErrors->GetBufferPointer();
			MessageBoxA(NULL, (const CHAR*)pCompilErrors, "Fx Compile Error", MB_OK|MB_ICONEXCLAMATION);
			softParticleFx= NULL;
		}
		MessageBoxA(NULL, "error", "Fx Compile Error", MB_OK|MB_ICONEXCLAMATION);
	}

	if (FAILED(D3DXCreateEffectFromFile(gd3dDevice, TEXT("sky.fx"), 0, 0, 0, 0, &skyFx, &pBufferErrors)))
	{
		if (pBufferErrors)
		{
			LPVOID pCompilErrors = pBufferErrors->GetBufferPointer();
			MessageBoxA(NULL, (const CHAR*)pCompilErrors, "Fx Compile Error", MB_OK|MB_ICONEXCLAMATION);
			skyFx= NULL;
		}
		MessageBoxA(NULL, "error", "Fx Compile Error", MB_OK|MB_ICONEXCLAMATION);
	}

	SAFE_RELEASE(pBufferErrors);

	UINT w, h;
	w=h= 512;
	HRESULT hr;
	//Creates a RT for the depth map:
	hr= gd3dDevice->CreateTexture(w,h, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &pDepthMap, NULL);
	hr= gd3dDevice->CreateDepthStencilSurface(w,h, D3DFMT_D24X8,D3DMULTISAMPLE_NONE,0, TRUE, &pDepthMapZ, NULL);
	pDepthMap->GetSurfaceLevel(0, &pDepthMapSurf);


	HR(D3DXCreateTextureFromFile(gd3dDevice, TEXT("media\\smoke.dds"), &particleMap));
	HR(D3DXCreateTextureFromFile(gd3dDevice, TEXT("media\\Sky1.jpg"), &skyMap));


	//A screen aligned quad needed to properly clear the depth map:
	D3DXVECTOR3 *Quad;
	hr=gd3dDevice->CreateVertexBuffer(sizeof(D3DXVECTOR3)*4, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &pVBQuad, NULL);
	hr=pVBQuad->Lock(0,0, (void**)&Quad, 0);

	Quad[0]= D3DXVECTOR3(-1.0f, 1.0f, 0.5f);
	Quad[1]= D3DXVECTOR3( 1.0f, 1.0f, 0.5f);
	Quad[2]= D3DXVECTOR3(-1.0f,-1.0f, 0.5f);
	Quad[3]= D3DXVECTOR3( 1.0f, -1.0f, 0.5f);

	hr=pVBQuad->Unlock();

	D3DVERTEXELEMENT9 decl[] =
    {
        { 0, 0,  D3DDECLTYPE_FLOAT3,   D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
        D3DDECL_END()
    };
	if (FAILED(gd3dDevice->CreateVertexDeclaration(decl, &pDeclPos)))exit(0);

	HR(LoadXFile("media\\skydome2.x", &mSkyMesh, mSkyMtrls, mSkyTextures))

	// The aspect ratio depends on the backbuffer dimensions, which can 
	// possibly change after a reset.  So rebuild the projection matrix.
	w = (float)md3dPP.BackBufferWidth;
	h = (float)md3dPP.BackBufferHeight;
	gCamera->setLens(D3DX_PI * 0.25f, w/h, 0.01f, 5000.0f);
	soft=true;
}




